"""A simple conversion of arbitrary feed data to JSON."""

import os
import json
import time
import hashlib
import feedparser

import settings

DEFAULT_FEED_ATTRS = ["title", "link", "entries"]
CACHE_DIR = os.path.join(settings.ROOT_DIR, 'feed_data')
RSS_URL = 'http://www.linaro.org/linaro-blog/category/patch-highlights/feed/'


class Feed2JSON(object):
    """A Simple way to convert a feed into json for storage."""

    def __init__(self, cache_dir=CACHE_DIR):
        """initialise the Feed2Json object"""
        self.cache_dir = cache_dir
        if not os.path.isdir(self.cache_dir):
            raise ValueError("Cache Directory %s is not a valid directory"
                             " path" % self.cache_dir)

    def get_hash_url(self, url):
        """Create a hash of the RSS url."""
        md5_ = hashlib.md5()
        md5_.update(url)
        return md5_.hexdigest()

    @staticmethod
    def to_json(python_object):
        """Custom serialisationto handle time.struct_time."""
        if isinstance(python_object, time.struct_time):
            return {'__class__': 'time.asctime',
                    '__value__': time.asctime(python_object)}

        raise TypeError(repr(python_object) + ' is not JSON serializable')

    @staticmethod
    def from_json(json_object):
        """Custom de-serialisation."""
        if '__class__' in json_object:
            if json_object['__class__'] == 'time.asctime':
                return time.strptime(json_object['__value__'])
        return json_object

    def get_feed(self, url):
        """Get the feed."""
        data = self.get_cached_data(url)
        etag = data.get("etag")
        last_modified = data.get("last_modified")
        fp = feedparser.parse(url, etag=etag, modified=last_modified)
        return fp, data

    def process_feed(self, url, feed_attrs=None):
        """Process the feed and put it into json"""

        if feed_attrs is None:
            feed_attrs = DEFAULT_FEED_ATTRS
        data = {}
        fp, data_ = self.get_feed(url)

        status = getattr(fp, "status", None)
        if not status or status != 200:
            return data_
        else:
            data["etag"] = getattr(fp, "etag", None)
            data["last_modified"] = getattr(fp, "modified", None)
            for attr in feed_attrs:
                data[attr] = getattr(fp, attr, None)

            hash_url = self.get_hash_url(url)
            cache_file = os.path.join(self.cache_dir, hash_url)

            with open(cache_file, "w") as fh:
                fh.write(json.dumps(data, default=self.to_json))

            return data

    def get_cached_data(self, url):
        """Get data from JSON cache."""
        data = {}
        hash_url = self.get_hash_url(url)
        cache_file = os.path.join(self.cache_dir, hash_url)
        if os.path.isfile(cache_file):
            with open(cache_file, "r") as fh:
                raw_data = fh.read()
                if raw_data:
                    data = json.loads(raw_data, object_hook=self.from_json)
        return data
