import calendar
from datetime import date, datetime, timedelta

from dateutil import relativedelta

from django.contrib.auth.models import User
from django.db import connection, models

from patchwork.models import Patch, Person, Project, State

# Patches for any of these projects are not included in our stats.
PROJECTS_TO_IGNORE_IN_STATS = [
    'other-unknown', 'linaro-kernel', 'gcc-linaro', 'gdb-linaro',
    'not-upstream', 'devicetree-discuss']

GERRIT_ROOT = 'https://review.source.android.com/'


class GerritChange(models.Model):
    # We use gerrit's ID as the primary key.
    gerrit_id = models.IntegerField(primary_key=True)
    last_updated_on = models.DateTimeField()
    # Gerrit's internal status.
    status = models.TextField(null=False, blank=False)
    change_id = models.TextField(unique=True, null=False)
    patch = models.ForeignKey(Patch)

    @property
    def url(self):
        return GERRIT_ROOT + str(self.gerrit_id)

    def save(self):
        from patchmetrics.gerrit import get_state_for_aosp_status
        patch_state = get_state_for_aosp_status(self.status)
        if self.patch.state != patch_state:
            self.patch.state = patch_state
            self.patch.save()
        super(GerritChange, self).save()


class Team(models.Model):
    name = models.CharField(max_length=255, null=False, unique=True)
    display_name = models.CharField(max_length=255, null=False)

    def add_member(self, member):
        return TeamMembership(team=self, member=member)

    @property
    def members(self):
        # Find out what are the alternative ways of spelling this.
        memberships = TeamMembership.objects.filter(team=self)
        return User.objects.filter(
            id__in=memberships.values('member_id'))

    def __str__(self):
        return self.display_name

    @property
    def patches(self):
        """All patches authored by any member of this team."""
        members = Person.objects.filter(user__in=self.members)
        return Patch.objects.filter(models.Q(author__in=members))

    def get_patch_count(self, start_date=None, end_date=None):
        """Return the number of non-superseded patches submitted by this team.

        Return a tuple with the total number of patches and the number of
        patches that have been accepted upstream.
        """
        patches = self.patches
        accepted = patches.filter(
            state=State.objects.get(name='Accepted'))
        if start_date is not None:
            patches = patches.filter(date__gte=start_date)
            accepted = accepted.filter(
                date_last_state_change__gte=start_date)
        if end_date is not None:
            patches = patches.filter(date__lte=end_date)
            accepted = accepted.filter(
                date_last_state_change__lte=end_date)
        return patches.count(), accepted.count()


class TeamMembership(models.Model):
    team = models.ForeignKey(Team, null=False)
    member = models.ForeignKey(User, null=False)

    def __str__(self):
        return "%s on %s" % (self.member.username, self.team.display_name)

    class Meta:
        unique_together = [('member', 'team')]


class LinaroPatchManager(models.Manager):

    def get_query_set(self):
        qs = super(LinaroPatchManager, self).get_query_set().filter(
            author__in=Person.objects.filter(email__iendswith='@linaro.org'))
        projects_to_skip = Project.objects.filter(
            linkname__in=PROJECTS_TO_IGNORE_IN_STATS)
        return qs.exclude(models.Q(state=State.objects.get(name='Superseded'))
                          | models.Q(project__in=projects_to_skip))


class LinaroPatch(Patch):
    """A Patch submitted by a Linaro engineer."""

    objects = LinaroPatchManager()

    class Meta:
        proxy = True


# January 2011 is when we started tracking patches.
PATCH_TRACKING_EPOCH = datetime(year=2011, month=1, day=1)


def get_patches_waiting_for_feedback_by_age_range(age_increment):
    """Return the distribution of pending patches grouped by age_increment.

    :param age_increment: Group patches with age (in days) in that range
        together. For example, for age_increment==30 patches with age between
        0 and 30 days would be grouped together.

    Return a sequence of elements of the form (index, count), where the first
    item represents the age range (e.g. for age_increment==30, 0 is for the
    0-30 range, 1 is for the 30-60 range, and so on) and the second is the
    number of patches with an age in that range.

    Only patches in the New state and authored by Linaro engineers are
    considered here.
    """
    cursor = connection.cursor()
    projects_to_skip = ",".join(
        "'%s'" % name for name in PROJECTS_TO_IGNORE_IN_STATS)
    sql = """
        SELECT cast(((
            extract(
                -- Number of days between the patch submission and now.
                EPOCH FROM age(
                    current_date,
                    date_trunc('day', date))) / 86400
            -- Cast to integer and divide by the age increment to group by age
            -- range rather than exact age.
            )) AS integer) / %(increment)d AS age,
            count(id)
        FROM patchwork_patch
        WHERE state_id = %(state)d
            -- Need to duplicate here the filtering done in LinaroPatch
            -- because we bypass django and use raw SQL.
            AND project_id NOT IN (
                SELECT id FROM patchwork_project
                WHERE linkname IN (%(projects_to_skip)s))
            AND author_id IN (
                SELECT id FROM patchwork_person
                WHERE email ILIKE '%(linaro)s')
        GROUP BY age;
        """ % dict(increment=age_increment,
                   state=State.objects.get(name='New').id,
                   linaro='%%@linaro.org',
                   projects_to_skip=projects_to_skip)
    cursor.execute(sql)
    return cursor.fetchall()


def get_patches_accepted_by_age_range(age_increment, team=None):
    """Return the distribution of accepted patches grouped by age_increment.

    :param age_increment: Group patches with age (in days) in that range
        together. For example, for age_increment==30 patches with age between
        0 and 30 days would be grouped together.
    :param team: Return only patches submitted by members of the given team.

    Return a sequence of elements of the form (index, count), where the first
    item represents the age range (e.g. for age_increment==30, 0 is for the
    0-30 range, 1 is for the 30-60 range, and so on) and the second is the
    number of patches with an age in that range.

    Only patches in the Accepted state and authored by Linaro engineers are
    considered here.
    """
    accepted = State.objects.get(name='Accepted')
    projects_to_skip = ",".join(
        "'%s'" % name for name in PROJECTS_TO_IGNORE_IN_STATS)
    replacements = dict(age_increment=age_increment, state=accepted.id,
                        projects_to_skip=projects_to_skip)
    cursor = connection.cursor()
    sql = """
        SELECT cast(((
            extract(
                -- Number of days between the patch submission and its
                -- acceptance.
                EPOCH FROM age(
                    date_trunc('day', date_last_state_change),
                    date_trunc('day', date))) / 86400
            -- Cast to integer and divide by the age increment to group by age
            -- range rather than exact age.
            )) AS integer) / %(age_increment)s AS age,
            count(id)
        FROM patchwork_patch
        WHERE date_last_state_change IS NOT NULL
              AND state_id = %(state)s
              -- Need to duplicate here the filtering done in LinaroPatch
              -- because we're bypassing django and using raw SQL.
              AND project_id NOT IN (
                  SELECT id FROM patchwork_project
                  WHERE linkname IN (%(projects_to_skip)s))
              AND author_id IN (
                  SELECT id FROM patchwork_person
                  WHERE email ILIKE '%%@linaro.org')
        """
    if team is not None:
        sql += """
            AND author_id IN (
                SELECT id FROM patchwork_person WHERE user_id IN (
                    SELECT member_id FROM patchmetrics_teammembership
                    WHERE team_id = %(team_id)s))
            """
        replacements['team_id'] = team.id


    sql += " GROUP BY age;"
    cursor.execute(sql, replacements)
    return cursor.fetchall()


def get_number_of_patches_accepted():
    patches = LinaroPatch.objects.filter(
        state=State.objects.get(name='Accepted'))
    return patches.count()


def get_number_of_patches_waiting_for_feedback():
    """Return number of patches in the 'New' state and the number of patches
    in the same state that are 30 or more days old.
    """
    patches = LinaroPatch.objects.filter(state=State.objects.get(name='New'))
    total_count = patches.count()
    more_than_30_days_old_count = patches.filter(
        date__lte=(datetime.now() - timedelta(days=30))).count()
    return total_count, more_than_30_days_old_count


def get_patches_submitted_and_accepted(start_date, end_date):
    """Return two lists of patches: submitted and accepted.

    The first list contains all patches submitted on the given period, in any
    state other than Superseded.

    The second list contains the patches that have been accepted on the given
    period, regardless of when they were submitted.
    """
    submitted = LinaroPatch.objects.filter(
        date__gte=start_date, date__lte=end_date)
    accepted = LinaroPatch.objects.filter(
        date_last_state_change__gte=start_date,
        date_last_state_change__lte=end_date,
        state=State.objects.get(name='Accepted'))
    return submitted, accepted


def get_distinct_author_count(year, month):
    """Return the number of authors of patches submitted on the given month.

    Only Linaro-submitted patches (as defined by the LinaroPatch proxy model)
    are considered here.
    """
    start_date = datetime(year=year, month=month, day=1)
    end_date = (start_date + relativedelta.relativedelta(months=1) -
                timedelta(seconds=1))
    patches = LinaroPatch.objects.filter(
        date__gte=start_date, date__lte=end_date)
    return patches.values('author').distinct().count()


def get_patches_per_month_up_to(year, month, count_function, limit=None):
    """Return two lists with patches submitted and accepted per month.

    Each list will include counts for the months between PATCH_TRACKING_EPOCH
    and the given month.

    Each element in those lists is of the form [timestamp, count], with the
    timestamp in miliseconds.

    If a limit is specified, return the counts for up to limit previous
    months.

    :param count_function: A callable which takes a start and end date and
        return a two-tuple with the number of submitted and accepted patches
        for that time period.
    """
    start_date = PATCH_TRACKING_EPOCH
    end_date = (start_date + relativedelta.relativedelta(months=1) -
                timedelta(seconds=1))
    accepted = []
    submitted = []
    current_date = date(year, month, 1)
    while start_date.date() <= current_date:
        month_timestamp = calendar.timegm(start_date.timetuple()) * 1000
        n_submitted, n_accepted = count_function(start_date, end_date)
        submitted.append([month_timestamp, n_submitted])
        accepted.append([month_timestamp, n_accepted])
        start_date = end_date + timedelta(seconds=1)
        end_date = (start_date + relativedelta.relativedelta(months=1) -
                    timedelta(seconds=1))
    while limit is not None and len(submitted) > limit:
        # Pop from the beginning as those as the oldest months.
        submitted.pop(0)
        accepted.pop(0)
    return submitted, accepted


def get_overall_patch_count(start_date, end_date):
    """Return the number of submitted/accepted patches for the given period.
    """
    submitted, accepted = get_patches_submitted_and_accepted(
        start_date, end_date)
    return submitted.count(), accepted.count()


def get_submitted_patch_counts_for_month(year, month):
    """Return the number of patches submitted and accepted on the given month.

    The return is a 2-tuple of integers where the first element is the number
    of non-Superseded patches submitted on the given month and the second is
    the number of patches that were submitted and accepted on the same month.
    """
    start_date = datetime(year=year, month=month, day=1)
    end_date = (start_date + relativedelta.relativedelta(months=1) -
                timedelta(seconds=1))
    submitted, accepted = get_patches_submitted_and_accepted(
        start_date, end_date)
    total = submitted.count()
    accepted = accepted.filter(date__gte=start_date, date__lte=end_date)
    return total, accepted.count()


def get_number_of_old_patches_accepted_on_month(year, month):
    """Return the number of patches submitted on previous months that were
    accepted on the given month.
    """
    start_date = datetime(year=year, month=month, day=1)
    end_date = (start_date + relativedelta.relativedelta(months=1) -
                timedelta(seconds=1))
    patches = LinaroPatch.objects.filter(
        date__lt=start_date, state=State.objects.get(name='Accepted'),
        date_last_state_change__gte=start_date,
        date_last_state_change__lte=end_date)
    return patches.count()


def get_oldest_patches_waiting_for_feedback():
    patches = LinaroPatch.objects.filter(state=State.objects.get(name='New'))
    return patches.order_by('date')[:10]


# This is a hack to make everybody able to edit all projects without having to
# change the Project class directly as that wouldn't be accepted upstream.
# XXX: Unfortunately, this hack interferes with the test suite, breaking 2
# tests that expect project.is_editable() to behave differently.  We could
# monkey patch those tests from here, but for now I'll leave them failing.
def project_is_editable(self, user):
    return user.is_authenticated()

Project.is_editable = project_is_editable
