from datetime import datetime, timedelta
from dateutil import relativedelta
import time

from django.test import TestCase

from patchwork.models import Patch, State, User
from patchmetrics import models
from patchmetrics.models import (
    GerritChange,
    get_distinct_author_count,
    get_overall_patch_count,
    get_patches_per_month_up_to,
    get_patches_accepted_by_age_range,
    get_patches_waiting_for_feedback_by_age_range,
    get_submitted_patch_counts_for_month,
    LinaroPatch, PROJECTS_TO_IGNORE_IN_STATS)
from patchmetrics.tests.factory import factory


class TestTeam(TestCase):

    def setUp(self):
        super(TestTeam, self).setUp()

    def test_basic(self):
        team = factory.makeTeam()
        self.assertEqual(team.display_name, str(team))

    def test_add_member(self):
        member = factory.makeUser()
        team = factory.makeTeam()
        membership = team.add_member(member)
        self.assertEqual(team, membership.team)
        self.assertEqual(member, membership.member)

    def test_members(self):
        membership = factory.makeTeamMembership()
        membership2 = factory.makeTeamMembership(team=membership.team)
        members = set([membership.member, membership2.member])
        self.assertEqual(members, set(membership.team.members))

    def test_patches(self):
        patch = factory.makePatch()
        team = self._create_team_with_member(patch.author)
        # This patch is not authored by a member of the team we created above,
        # so it doesn't show up on the team's patch list.
        patch2 = factory.makePatch()
        self.assertEqual([patch], list(team.patches))

    def test_get_patch_count_for_date(self):
        recent_patch = factory.makePatch(
            date=(datetime.now() - timedelta(days=5)))
        member = recent_patch.author
        team = self._create_team_with_member(member)
        old_patch = factory.makePatch(
            author=member, date=(datetime.now() - timedelta(days=60)))
        fresh_patch = factory.makePatch(
            author=member, date=datetime.now())
        a_month_ago = datetime.now() - timedelta(days=30)
        two_days_ago = datetime.now() - timedelta(days=2)
        patch_count = team.get_patch_count(
            start_date=a_month_ago, end_date=two_days_ago)
        self.assertEqual((1, 0), patch_count)

    def test_get_patch_count(self):
        patch = factory.makePatch()
        team = self._create_team_with_member(patch.author)
        self.assertEqual((1, 0), team.get_patch_count())

    def _create_team_with_member(self, person):
        user = person.user
        if user is None:
            user = User.objects.create_user(
                person.name, person.email, password=None)
        user.save()
        team = factory.makeTeam()
        membership = team.add_member(user)
        membership.save()
        return team


class TestQueryFunctions(TestCase):

    def test_get_distinct_author_count(self):
        patch1 = factory.makeLinaroPatch()
        # This is just to create a Linaro patch using the same author as the
        # previous patch. We don't add an extra 'author' argument to
        # makeLinaroPatch() because that would allow people to inadvertently
        # create a non-Linaro patch using it.
        factory.makePatch(author=patch1.author)
        factory.makeLinaroPatch()
        now = datetime.now()
        last_month = now - relativedelta.relativedelta(months=1)
        factory.makeLinaroPatch(date=last_month)
        # We have 3 patches submitted this month, from two different authors,
        # and one patch submitted the previous month, so the number of
        # distinct authors this month is 2.
        self.assertEqual(2, get_distinct_author_count(now.year, now.month))


class TestChartHelpers(TestCase):

    def test_get_submitted_patch_counts_for_month(self):
        patch1 = factory.makeLinaroPatch()
        patch2 = factory.makeLinaroPatch(state_name='Accepted')
        now = datetime.now()
        last_month = now - relativedelta.relativedelta(months=1)
        patch3 = factory.makeLinaroPatch(date=last_month)
        # Here we have two patches submitted this month (patch1 and patch2),
        # with one of them also accepted this month (patch2). The other one
        # (patch3) was submitted last month, so it's not considered here.
        self.assertEqual(
            (2, 1), get_submitted_patch_counts_for_month(now.year, now.month))

    def test_get_patches_per_month_up_to(self):
        patch1 = factory.makeLinaroPatch()
        patch2 = factory.makeLinaroPatch(state_name='Accepted')
        now = datetime.now()
        last_month = now - relativedelta.relativedelta(months=1)
        patch3 = factory.makeLinaroPatch(date=last_month)
        # Although we have a patch from 3 months ago we won't get a count for
        # that month because we'll ask get_overall_patches_per_month_up_to to
        # give us the counts for the last two months (this one and the
        # previous) only.
        three_months_ago = now - relativedelta.relativedelta(months=3)
        patch4 = factory.makeLinaroPatch(date=three_months_ago)
        # Monkey-patch PATCH_TRACKING_EPOCH to make the results predictable.
        orig_epoch = models.PATCH_TRACKING_EPOCH
        models.PATCH_TRACKING_EPOCH = datetime(
            year=last_month.year, month=three_months_ago.month, day=1)
        try:
            submitted, accepted = get_patches_per_month_up_to(
                now.year, now.month, get_overall_patch_count, limit=2)
        finally:
            models.PATCH_TRACKING_EPOCH = orig_epoch

        # get_overall_patches_per_month_up_to() returns timestamps in
        # miliseconds so we need to divide them by 1000 here.
        submitted = [(time.gmtime(timestamp / 1000).tm_mon, count)
                     for timestamp, count in submitted]
        self.assertEqual([(last_month.month, 1), (now.month, 2)], submitted)

        accepted = [(time.gmtime(timestamp / 1000).tm_mon, count)
                    for timestamp, count in accepted]
        self.assertEqual([(last_month.month, 0), (now.month, 1)], accepted)

    def test_get_patches_accepted_by_age_range(self):
        four_days_ago = datetime.now() - timedelta(days=4)
        patch1 = factory.makeLinaroPatch(state_name='Accepted')
        patch2 = factory.makeLinaroPatch(
            date=four_days_ago, state_name='Accepted')
        patch3 = factory.makePatch()
        patch4 = factory.makeLinaroPatch(state_name='New')
        age_increment = 3
        # Our first patch has today as its date so it's in the first group;
        # the second was created 4 days ago and ends up in the second group;
        # the third is not authored by a Linaro engineer so is ignored; the
        # fourth is authored by a Linaro engineer but is not in the Accepted
        # state so is also ignored.
        self.assertEqual([(0, 1L), (1, 1L)],
                         get_patches_accepted_by_age_range(age_increment))

    def test_get_patches_accepted_by_age_range_for_team(self):
        four_days_ago = datetime.now() - timedelta(days=4)
        patch1 = factory.makeLinaroPatch(state_name='Accepted')
        patch2 = factory.makeLinaroPatch(
            date=four_days_ago, state_name='Accepted')
        author = patch1.author
        author.user = factory.makeUser()
        author.save()
        membership = factory.makeTeamMembership(member=author.user)
        team = membership.team
        patch2.author = author
        patch2.save()
        patch3 = factory.makeLinaroPatch(
            date=four_days_ago, state_name='Accepted')
        age_increment = 3
        # The first two patches are authored by a member of the team we're
        # interested in, so they are included in the counts, whereas the third
        # one is not so it is not included.
        self.assertEqual([(0, 1L), (1, 1L)],
                         get_patches_accepted_by_age_range(
                            age_increment, team))

    def test_get_patches_waiting_for_feedback_by_age_range(self):
        seven_days_ago = datetime.now() - timedelta(days=7)
        patch1 = factory.makeLinaroPatch(state_name='New')
        patch2 = factory.makeLinaroPatch(
            date=seven_days_ago, state_name='New')
        patch3 = factory.makePatch()
        patch4 = factory.makeLinaroPatch(state_name='Accepted')
        age_increment = 3
        # Our first patch has today as its date so it's in the first group;
        # the second was created seven days ago and ends up in the third
        # group; the third is not authored by a Linaro engineer so is ignored;
        # the fourth is authored by a Linaro engineer but is not in the New
        # state so is also ignored.
        self.assertEqual(
            [(0, 1L), (2, 1L)],
            get_patches_waiting_for_feedback_by_age_range(age_increment))


class TestLinaroPatch(TestCase):

    def test_ingored_projects(self):
        # Patches for any of our ignored projects are not included in
        # LinaroPatch.objects.
        ignored_project = factory.makeProject(
            linkname=PROJECTS_TO_IGNORE_IN_STATS[0])
        patch1 = factory.makeLinaroPatch(project=ignored_project)
        patch2 = factory.makeLinaroPatch()
        self.assertEquals([patch1, patch2], list(Patch.objects.all()))
        self.assertEquals([patch2.id],
                          [patch.id for patch in LinaroPatch.objects.all()])

    def test_superseded_patches(self):
        # Patches in the Superseded state are not included in
        # LinaroPatch.objects.
        patch = factory.makeLinaroPatch()
        self.assertEquals(1, len(LinaroPatch.objects.all()))
        patch.state = State.objects.get(name='Superseded')
        patch.save()
        self.assertEquals(0, len(LinaroPatch.objects.all()))

    def test_non_linaro_patches(self):
        # Patches submitted by people not using @linaro.org addresses are
        # never included in LinaroPatch.objects.
        patch_submitted_by_linaro_engineer = factory.makeLinaroPatch()
        patch_submitted_by_non_linaro_engineer = factory.makePatch()
        patches = LinaroPatch.objects.all()
        self.assertEquals(1, len(patches))
        self.assertEquals(
            patch_submitted_by_linaro_engineer.id, patches[0].id)


class TestPatchCounts(TestCase):

    def test_get_patch_count_for_date(self):
        project = factory.makeProject()
        recent_patch = factory.makeLinaroPatch(
            project=project, date=(datetime.now() - timedelta(days=5)))
        old_patch = factory.makeLinaroPatch(
            project=project, date=(datetime.now() - timedelta(days=60)))
        fresh_patch = factory.makeLinaroPatch(
            project=project, date=datetime.now())
        a_month_ago = datetime.now() - timedelta(days=30)
        two_days_ago = datetime.now() - timedelta(days=2)
        patch_count = project.get_patch_count(
            start_date=a_month_ago, end_date=two_days_ago)
        self.assertEqual((1, 0), patch_count)

    def test_get_patch_count(self):
        patch = factory.makeLinaroPatch()
        self.assertEqual((1, 0), patch.project.get_patch_count())


class TestGerritChange(TestCase):

    def test_save_does_not_touch_patch_state_when_not_needed(self):
        patch = factory.makePatch()
        self.assertEqual(State.objects.get(name='New'), patch.state)
        change = GerritChange(
            gerrit_id=1, patch=patch, status='NEW', change_id='foo',
            last_updated_on=datetime.now())
        change.save()
        self.assertEqual(State.objects.get(name='New'), patch.state)

    def test_save_updates_patch_state_when_needed(self):
        patch = factory.makePatch()
        self.assertEqual(State.objects.get(name='New'), patch.state)
        change = GerritChange(
            gerrit_id=1, patch=patch, status='MERGED', change_id='foo',
            last_updated_on=datetime.now())
        change.save()
        self.assertEqual(State.objects.get(name='Accepted'), patch.state)


class TestProject(TestCase):

    def test_not_editable_by_unauthenticated_user(self):
        project = factory.makeProject()
        user = factory.makeUser()
        user.is_authenticated = lambda: False
        self.assertFalse(project.is_editable(user))

    def test_is_editable_by_any_authenticated_user(self):
        project = factory.makeProject()
        user = factory.makeUser()
        user.is_authenticated = lambda: True
        self.assertTrue(project.is_editable(user))
