from copy import copy
from datetime import datetime

from django.test import TestCase

from patchmetrics.gerrit_values import (
    GERRIT_QUERY_FORMAT,
    GERRIT_QUERY_OWNER,
    GERRIT_QUERY_PG_SIZE,
    GERRIT_QUERY_RESUME_SORTKEY,
)

from patchmetrics.gerrit import (
    CHANGES_QUERY_PATH,
    QUERY_STRING_START,
    ROOT_REST_URL,
    create_gerrit_rest_query,
    create_patch_from_aosp_change,
    create_request_query_url,
    get_project_for_aosp_name,
    get_project_name_for_aosp_name,
    update_local_gerrit_change_from_aosp_change,
    )
from patchmetrics.tests.factory import factory


class TestGerritHelpers(TestCase):

    aosp_bionic = 'platform/bionic'
    date_created = datetime(2011, 01, 01)
    change_dict = dict(
        id=12345, status='NEW', project=aosp_bionic, subject='Foo bar',
        date_created=date_created, last_updated_on=datetime.now(),
        change_id='I098ab432')

    def test_get_project_for_aosp_name_with_existing_project(self):
        aosp_name = 'kernel/tegra'
        name = get_project_name_for_aosp_name(aosp_name)
        project = factory.makeProject(linkname=name)
        self.assertEqual(project, get_project_for_aosp_name(aosp_name))

    def test_get_project_for_aosp_name_non_existing_project(self):
        aosp_name = 'toolchain/build'
        project = get_project_for_aosp_name(aosp_name)
        self.assertEqual(
            get_project_name_for_aosp_name(aosp_name), project.linkname)

    def test_create_patch_from_aosp_change(self):
        author = factory.makePerson()
        patch = create_patch_from_aosp_change(self.change_dict, author)
        self.assertTrue(patch.gerrit_change is not None)
        self.assertTrue(author, patch.author)
        self.assertTrue(author, patch.submitter)
        self.assertEqual(self.change_dict['subject'], patch.name)
        self.assertEqual('New', patch.state.name)
        self.assertTrue(patch.date_last_state_change is None)
        self.assertEqual(get_project_name_for_aosp_name(self.aosp_bionic),
                         patch.project.linkname)

    def test_create_patch_from_aosp_change_with_state_change(self):
        # Here the gerrit change is not in the new state, so the new patch
        # will be created with the appropriate state and
        # date_last_state_change set to last_updated_on.
        change_dict = copy(self.change_dict)
        change_dict['status'] = 'MERGED'
        patch = create_patch_from_aosp_change(
            change_dict, factory.makePerson())
        self.assertEqual('Accepted', patch.state.name)
        self.assertEqual(
            change_dict['last_updated_on'], patch.date_last_state_change)

    def test_update_local_gerrit_change_from_aosp_change(self):
        gerrit_change = factory.makeGerritChange()
        change_dict = copy(self.change_dict)
        change_dict['id'] = gerrit_change.gerrit_id
        change_dict['status'] = 'MERGED'
        update_local_gerrit_change_from_aosp_change(
            gerrit_change, change_dict)
        self.assertEqual('MERGED', gerrit_change.status)
        self.assertEqual('Accepted', gerrit_change.patch.state.name)
        self.assertEqual(change_dict['subject'], gerrit_change.patch.name)

    def test_create_gerrit_rest_query(self):
        """Tests the creation of a valid gerrit query."""
        query_params = {'key1': 'value1', 'key2': 'value2'}
        self.assertEqual('&q=key2:value2+key1:value1',
                         create_gerrit_rest_query(query_params))

    def test_create_gerrit_rest_query_empty(self):
        """Tests gerrit query creation with empty dictionary."""
        self.assertEqual('', create_gerrit_rest_query({}))

    def test_create_gerrit_rest_query_with_none(self):
        """Tests that if a value is None it is not added."""
        query_params = {'key1': 'value1', 'key2': None}
        self.assertEqual('&q=key1:value1',
                         create_gerrit_rest_query(query_params))

    def test_create_gerrit_request_query_with_list(self):
        """Tests that the function accepts only a dictionary, passing a list.
        """
        self.assertRaises(TypeError, create_gerrit_rest_query, [])

    def test_create_gerrit_request_query_with_tuple(self):
        """Tests that the function accepts only a dictionary, passing a tuple.
        """
        self.assertRaises(TypeError, create_gerrit_rest_query, ())

    def test_create_request_query_url(self):
        """Tests the creation of a valid request URL."""
        params = {'format': 'JSON', 'n': 100}
        query_params = {'owner': 'person@example.org', 'sortkey_after': 'abcd'}
        expected_result = (ROOT_REST_URL + CHANGES_QUERY_PATH +
                            QUERY_STRING_START + "n=100&format=JSON&q=owner:"
                            "%22person%40example.org%22+resume_sortkey:abcd")
        self.assertEquals(expected_result,
                          create_request_query_url(ROOT_REST_URL,
                                                   CHANGES_QUERY_PATH,
                                                   params, query_params))

    def test_create_request_query_url_quoted(self):
        """Tests the creation of a valid request URL with a complex email."""
        params = {GERRIT_QUERY_FORMAT: 'JSON', GERRIT_QUERY_PG_SIZE: 100}
        query_params = {GERRIT_QUERY_OWNER: 'name+surname.person@example.org',
                        GERRIT_QUERY_RESUME_SORTKEY: 'abcd'}
        expected_result = (ROOT_REST_URL + CHANGES_QUERY_PATH +
                            QUERY_STRING_START + "n=100&format=JSON&q=owner:"
                            "%22name%2Bsurname.person%40example.org%22+"
                            "resume_sortkey:abcd")
        self.assertEquals(expected_result,
                          create_request_query_url(ROOT_REST_URL,
                                                   CHANGES_QUERY_PATH,
                                                   params, query_params))
