import operator

from django.test import TestCase

from patchwork.models import Person
from patchmetrics.utils import (
    is_member,
    sync_memberships,
    get_sub_teams,
    sync_launchpad_memberships_for,
    )
from patchmetrics.tests.factory import PatchMetricsObjectFactory


class FakeTeam(object):
    def __init__(self, name=None, subteams=[]):
        self.name = name
        self.sub_teams = []
        for name in subteams:
            team = FakeTeam(name=name)
            self.sub_teams.append(team)


class FakePerson(object):

    def __init__(self, teams, email="a@b.com"):
        self.super_teams = []
        self.email = email
        for name in teams:
            team = FakeTeam(name=name)
            self.super_teams.append(team)


class TestSyncMembership(TestCase):

    def setUp(self):
        super(TestSyncMembership, self).setUp()
        self.factory = PatchMetricsObjectFactory()

    def test_get_sub_teams(self):
        expected = sorted(['infra', 'kernel', 'android'])
        linaro = FakeTeam(subteams=expected)
        subteams = get_sub_teams(linaro)
        self.assertEqual(expected, sorted(subteams.keys()))
        self.assertEqual(
            expected, sorted([team.name for team in subteams.values()]))

    def test_get_sub_teams_skips_some_teams(self):
        expected = sorted(['infra', 'kernel', 'android'])
        teams_to_skip = [
            'linaro-project-management',
            'linaro-oceo',
            'linaro-oevp',]
        input = expected + teams_to_skip
        linaro = FakeTeam(subteams=input)
        subteams = get_sub_teams(linaro)
        self.assertEqual(expected, sorted(subteams.keys()))
        self.assertEqual(
            expected, sorted([team.name for team in subteams.values()]))

    def test_is_member(self):
        person = FakePerson(['linaro', 'infra', 'android'])
        self.assertTrue(
            is_member(person, dict(multimedia=1, linaro=0)))
        self.assertFalse(
            is_member(person, ['multimedia', 'kernel']))

    def test_sync_launchpad_memberships_for(self):
        from patchmetrics import utils
        linaro = FakeTeam("linaro", ["subteam"])
        subteam = FakeTeam("subteam")
        emails = sorted(['foo@example.com', 'baz@example.com'])

        def get_lp_person_by_email(lp, email):
            return FakePerson(['subteam'], email)

        original_get_lp_person_by_email = utils.get_lp_person_by_email
        utils.get_lp_person_by_email = get_lp_person_by_email

        def sync_memberships(memberships):
            """fake sync_memberships used to test membeships.

            We don't check that FakeTeam and FakePerson objects match
            because they are often created in a way that is difficult
            to test, so we test some unique characteristics instead.
            """

            # We have only created 1 subteam
            self.assertEqual(len(memberships.keys()), 1)

            # The subteam should have the correct name
            team = memberships.keys()[0]
            self.assertEqual(team.name, subteam.name)

            # Now collect all email addresses stored in memberships
            found_emails = []
            for member in memberships[team].keys():
                found_emails += memberships[team][member]

            self.assertEqual(sorted(found_emails), emails)

        original_sync_memberships = utils.sync_memberships
        utils.sync_memberships = sync_memberships

        def sync_openid_urls():
            pass

        original_sync_openid_urls = utils.sync_openid_urls
        utils.sync_openid_urls = sync_openid_urls

        try:
            sync_launchpad_memberships_for(emails, linaro, None)

        finally:
            utils.get_lp_person_by_email = original_get_lp_person_by_email
            utils.sync_memberships = original_sync_memberships
            utils.sync_openid_urls = original_sync_openid_urls

    def test_sync_new_team(self):
        person = self.factory.makePerson()
        emails = sorted([person.email, 'new@example.com'])
        memberships = {
            # This will cause the creation of a new team, a new user (as
            # a member of the team) and a new person for the unseen email
            # address. The new user will also be linked to the existing
            # person.
            FakeTeamOrMember(name='non-existing-team'): {
                FakeTeamOrMember(person.name, person.name): emails}}

        created_memberships = sync_memberships(memberships)
        self.assertEqual(1, len(created_memberships))
        [membership] = created_memberships
        self.assertEqual('non-existing-team', membership.team.name)
        self.assertEqual(person.user, membership.member)

        people = Person.objects.filter(user=membership.member)
        self.assertEqual(emails, sorted([person.email for person in people]))

    def test_sync_just_add_members(self):
        team = self.factory.makeTeam()
        emails = sorted(['name1@example.com', 'random@example.com'])
        memberships = {
            # This will cause two persons (one for each email address) to
            # be created as well as a user (linking them together), which is
            # added as a member of the team created above.
            FakeTeamOrMember(name=team.name): {
                FakeTeamOrMember('random-name1'): emails}}

        created_memberships = sync_memberships(memberships)
        self.assertEqual(1, len(created_memberships))
        [membership] = created_memberships
        self.assertEqual(team.name, membership.team.name)
        self.assertEqual('name1@example.com', membership.member.email)

        people = Person.objects.filter(user=membership.member)
        self.assertEqual(emails, sorted([person.email for person in people]))

    def test_sync_nothing_to_create(self):
        membership = self.factory.makeTeamMembership()
        member = membership.member
        memberships = {
            # This won't cause anything to be created as the membership
            # already exists in our DB.
            membership.team: {
                FakeTeamOrMember(member.username): [member.email]}}

        created_memberships = sync_memberships(memberships)

        self.assertEqual([], list(created_memberships))

    def test_sync_name_fixme(self):
        person = self.factory.makePerson(is_user=False)
        team = self.factory.makeTeam()

        emails = [self.factory.getUniqueEmailAddress()]
        memberships = {
            # This will cause the creation of a Person for the unseen email
            # address in 'emails', a user account linked to this new person
            # entry and to the existing one (which has the same name as the
            # new), as well as a team membership for the new user on the team
            # created above.
            team: {FakeTeamOrMember(person.name, person.name): emails}}

        created_memberships = sync_memberships(memberships)
        self.assertEqual(1, len(created_memberships))
        [membership] = created_memberships
        self.assertEqual(team.name, membership.team.name)
        self.assertEqual(emails[0], membership.member.email)

        people = Person.objects.filter(user=membership.member)
        all_emails = sorted(emails + [person.email])
        self.assertEqual(
            all_emails, sorted([person.email for person in people]))


class FakeTeamOrMember(object):

    def __init__(self, name, display_name='Random name'):
        self.name = name
        self.display_name = display_name
