import logging
import time
import urllib2
import re

from patchmetrics.crowd import (
    CrowdException,
    CrowdNotFoundException,
)
from django.contrib.auth.models import User
from django.db.models import Q
from django_openid_auth.models import UserOpenID
from lazr.restfulclient.errors import HTTPError
from patchmetrics.models import Team
from patchwork.models import Person
from xml.dom.minidom import parseString

logging.basicConfig()
logger = logging.getLogger()

# Time in seconds to wait before performing another HTTP request.
REQUEST_WAIT = 30
# Max retries if something goes wrong, but it can be recoverd just retrying.
MAX_RETRIES = 3


def create_team_display_name(team):
    """Very simple and hackish way to create a display name for a team.

    :param team: The team name.
    """
    # XXX: since Crowd does not expose the displayName of a team, we hack one.
    parts = re.split("-|_", team)
    display_name = " ".join(parts)
    return display_name.title()


def sync_crowd_memberships(memberships):
    """Make sure the given memberships are represented in the database.

    :param memberships: A dict mapping each team to a list of members. Each
        member is in turn represented by a CrowdUser object.
        For example: {Team: [CrowdUser1, CrowdUser2, ...]}
    """
    created_memberships = []
    for team, members in memberships.iteritems():
        team_display_name = create_team_display_name(team)
        try:
            db_team = Team.objects.get(name=team)
        except Team.DoesNotExist:
            logger.info(
                'Creating new team: {0} ({1})'.format(team, team_display_name))
            db_team = Team(name=team, display_name=team_display_name)
            db_team.save()

        for member in members:
            people = []
            for email in member.emails:
                try:
                    person = Person.objects.get(email=email)
                except Person.DoesNotExist:
                    logger.info("Creating new person: {0} "
                                "({1})".format(member.name, email))
                    person = Person(name=member.display_name, email=email)
                    person.save()
                people.append(person)

            # Get all Person entries that might represent other email
            # addresses of this same user.
            people.extend(Person.objects.filter(name=member.display_name))

            user = get_user(member.name, member.emails)
            if user is None:
                user = User.objects.create_user(
                    member.name, member.emails[0], password=None)

            # Now link all the Person entries to the user account.
            for person in people:
                person.user = user
                person.save()

            # And finally, make sure the user is a member of the team.
            if user not in db_team.members:
                logger.info("Adding {0} as a member "
                            "of {1}".format(member.name, db_team.name))
                membership = db_team.add_member(user)
                membership.save()
                created_memberships.append(membership)

    return created_memberships


def sync_memberships(memberships):
    """Make sure the given memberships are represented in our database.

    :param memberships: A dict mapping teams to its members. Each member is in
        turn represented by a dict mapping its name to all its email
        addresses. e.g.:

        {Team: {Person: ['member1@example.com', ...], 'member2': [...]}}
    """
    created_memberships = []
    for team, members in memberships.items():
        try:
            db_team = Team.objects.get(name=team.name)
        except Team.DoesNotExist:
            logger.info(
                'Creating new team: %s (%s)' % (team.name, team.display_name))
            db_team = Team(name=team.name, display_name=team.display_name)
            db_team.save()

        for member, emails in members.items():
            display_name = member.display_name
            people = []
            # Make sure there's one Person entry for every email address.
            for email in emails:
                try:
                    person = Person.objects.get(email=email)
                except Person.DoesNotExist:
                    logger.info(
                        'Creating new person: %s (%s)' % (member.name, email))
                    person = Person(name=display_name, email=email)
                    person.save()
                people.append(person)

            # Get all Person entries that might represent other email
            # addresses of this same user.
            people.extend(Person.objects.filter(name=display_name))

            user = get_user(member.name, emails)
            if user is None:
                user = User.objects.create_user(
                    member.name, emails[0], password=None)

            # Now link all the Person entries to the user account.
            for person in people:
                person.user = user
                person.save()

            # And finally, make sure the user is a member of the team.
            if user not in db_team.members:
                logger.info('Adding %s as a member of %s)'
                            % (member.name, db_team.name))
                membership = db_team.add_member(user)
                membership.save()
                created_memberships.append(membership)

    return created_memberships


def get_sub_teams(parent_team):
    """Get all subteams of the given team from Launchpad.

    This function is not recursive, so teams that are more than one level of
    indirection down will not be picked up. Obviously this can be fixed if
    it is a problem.
    """

    teams_to_skip = [
        'linaro-project-management',
        'linaro-oceo',
        'linaro-oevp']

    teams = {}
    for team in parent_team.sub_teams:
        if team.name in teams_to_skip:
            continue

        teams[team.name] = team

    return teams


def get_lp_person_by_email(lp, email):
    return lp.people.getByEmail(email=email)


def is_member(person, teams):
    """Return the teams the given person is a member of."""

    # XXX: The .super_teams property was made for teams, so there's no
    # guarantee it wil always work for Persons as well.  Need to figure out
    # whether this is the appropriate way to get all teams a person is a
    # member of.
    member_of = []
    for team in person.super_teams:
        team_name = team.name
        if team_name in teams:
            member_of.append(teams[team_name])

    return member_of


def get_user(name, emails):
    """Return the user linked to the person with one of the given emails.

    If there are no users linked to a person with any of the given emails,
    return None.
    """
    query = (Q(person__in=Person.objects.filter(email__in=emails))
             | Q(username=name))
    users = User.objects.filter(query).distinct()
    if users.count() == 1:
        return users[0]
    elif users.count() > 1:
        logger.info("Found more than one user for {0}; "
                    "using the first one".format(emails))
        return users[0]
    else:
        return None


def sync_user_memberships(email_addresses, crowd, whitelisted_groups=[]):
    """If an input email matches one from Linaro Login, add them to the db.

    If an email address matches a user in Linaro Login, takes the CrowdUser
    object and passes it to sync_crowd_memberships, where CrowdUser are
    matched with User objects (many CrowdUser can map to a single User).
    The tail call to `sync_openid_urls` then adds OpenID URLs to each User
    object.

    :param: email_addresses: List of email addresses to look in Linaro Login.
    :type list
    :param crowd: The Crowd object instance to perform query to Linaro Login.
    :type Crowd
    :param whitelisted_groups: A list of valid groups/teams to create.
    :type list
    """
    memberships = {}
    for email in email_addresses:
        email = email.lower()

        user = None
        try:
            user = crowd.get_user_with_groups(email)
        except CrowdNotFoundException:
            # If there is not a user matching that email addess in Linaro,
            # move on.
            pass
        except CrowdException:
            # If something else went wrong, report it.
            logger.warning("Error while searching email address "
                           "'{0}'.".format(email))

        # No user with that email, or user is not part of any team.
        if user is None or not user.teams:
            continue
        else:
            for team in user.teams:
                # If we have a list of valid groups, obey it. Otherwise, all
                # groups are valid.
                if whitelisted_groups:
                    if team in whitelisted_groups:
                        if not team in memberships:
                            memberships[team] = []
                        memberships[team].append(user)
                else:
                    if not team in memberships:
                        memberships[team] = []
                    memberships[team].append(user)

    sync_crowd_memberships(memberships)


def sync_launchpad_memberships_for(input_email_addresses, team, lp):
    """If an input email matches a LP user, add them to the database.

    If an email address matches a user in Launchpad, takes the LP Person
    object and passes it, along with the matching email address and passes it
    to sync_memberships, where Persons are matched up with User objects (many
    Persons can map to a single User). sync_openid_urls then adds OpenID URLs
    to each User object.
    """
    teams = get_sub_teams(team)

    memberships = {}
    for email in input_email_addresses:
        email = email.lower()

        try:
            person = get_lp_person_by_email(lp, email)
        except HTTPError:
            # FIXME We should not encounter bad email addresses.
            # https://bugs.launchpad.net/linaro-patchmetrics/+bug/967290
            person = None  # Launchpad lookup failed.
            logger.warning("Launchpad rejected email lookup and returned "
                           "HTTPError. Offending email address: %s" % email)

        if not person:
            # No Launchpad entry for this person.
            continue

        # Counter for the retries to perform in case of an HTTP error 503.
        retries_count = 0
        while True:
            try:
                member_of = is_member(person, teams)
                for team in member_of:
                    if not team in memberships:
                        memberships[team] = {}

                    memberships[team][person] = [email]
            except HTTPError, ex:
                # We should not get an error 503 from Launchpad.
                # https://bugs.launchpad.net/linaro-patchmetrics/+bug/1012538
                if ex.status == 503 and retries_count < MAX_RETRIES:
                    retries_count += 1
                    time.sleep(REQUEST_WAIT)
                    continue
                else:
                    raise(ex)
            break

    sync_memberships(memberships)
    sync_openid_urls()


def sync_openid_urls():
    """Insert OpenID URLs for users missing them

    When sync_memberships creates a User it doesn't do this, so this function
    just patches up the database with that extra information.
    """
    # TODO
    # Crowd openid urls are like this:
    # https://login.linaro.org:8443/openidserver/users/email@linaro.org
    for user in User.objects.all():
        if UserOpenID.objects.filter(user=user).count() == 0:
            try:
                xrds = urllib2.urlopen(
                    'https://launchpad.net/~%s/+xrds' % user.username).read()
            except urllib2.HTTPError, e:
                if e.getcode() == 404:
                    # User doesn't exist in LP, so not much for us to do.
                    continue
                raise
            dom = parseString(xrds)
            identity = dom.getElementsByTagName(
                'LocalID')[0].childNodes[0].data
            if UserOpenID.objects.filter(claimed_id=identity).count() == 0:
                print 'Setting openid identity for %s' % user.username
                openid_identity = UserOpenID(
                    user=user, claimed_id=identity, display_id=identity)
                openid_identity.save()
