import calendar
from datetime import datetime
from time import mktime

from dateutil import relativedelta

from django.db import models
from django.shortcuts import get_object_or_404, render_to_response

from patchwork.requestcontext import PatchworkRequestContext
from patchwork.models import Patch, Person, Project, State
from patchmetrics.feed2json import Feed2JSON, RSS_URL
from patchmetrics.models import (
    get_distinct_author_count,
    get_number_of_old_patches_accepted_on_month,
    get_number_of_patches_accepted,
    get_number_of_patches_waiting_for_feedback,
    get_patches_accepted_by_age_range,
    get_patches_per_month_up_to,
    get_patches_waiting_for_feedback_by_age_range,
    get_submitted_patch_counts_for_month,
    get_oldest_patches_waiting_for_feedback,
    get_overall_patch_count,
    LinaroPatch,
    PROJECTS_TO_IGNORE_IN_STATS,
    Team)
from patchwork.views import generic_list
from patchwork.views.patch import patch as patchwork_patch_view


def patch(request, patch_id):
    # This is just so that we can access patches via /NNNN without a redirect,
    # in order for people to be able to easily lookup a patch given its ID.
    return patchwork_patch_view(request, patch_id)


def project(request, project_id):
    project = get_object_or_404(Project, linkname=project_id)
    context = generic_list(
        request, project, 'patchmetrics.views.project',
        view_args={'project_id': project.linkname},
        should_show_move_form=True)
    now = datetime.now()

    total, accepted = project.get_patch_count()
    context['project'] = project
    context['total_patches'] = total
    context['accepted_patches'] = accepted
    context['patches_per_month_chart'] = get_patches_per_month_chart(
        project, now.year, now.month)

    return render_to_response('patchmetrics/project.html', context)


def team(request, team_id):
    team = get_object_or_404(Team, name=team_id)
    member_names = []
    for member in team.members:
        profile = member.get_profile()
        name = profile.person().name
        if not name:
            name = profile.name()
        member_names.append(name)

    patches = team.patches.exclude(state=State.objects.get(name='Superseded'))
    # XXX: This is a hack because generic_list() is not meant to be used for
    # cross-project patch lists.
    context = generic_list(
        request, None, 'patchmetrics.views.team',
        view_args={'team_id': team.name}, patches=patches)

    now = datetime.now()
    submitted, accepted = team.get_patch_count()
    context['total_submitted'] = submitted
    context['total_accepted'] = accepted
    context['team'] = team
    context['member_names'] = sorted(member_names)
    context['patches_per_month_chart'] = get_patches_per_month_chart(
        team, now.year, now.month)
    age_increment = 7
    accepted = get_patches_accepted_by_age_range(
        age_increment=age_increment, team=team)
    context['time_to_acceptance_chart'] = (
        get_patch_age_distribution_chart(
            accepted, age_increment, 'timeToAcceptanceChart',
            color="'#89b529'"))
    return render_to_response('patchmetrics/team.html', context)


def overall_teams_report(request):
    context = PatchworkRequestContext(request)
    one_month_ago = datetime.now() - relativedelta.relativedelta(months=1)
    teams_data = []
    months = []
    for team in Team.objects.all().order_by('display_name'):
        submitted, accepted = get_patches_per_month_up_to(
            one_month_ago.year, one_month_ago.month, team.get_patch_count,
            limit=8)
        if len(months) == 0:
            for item in submitted:
                # The timestamps returned by get_patches_per_month_up_to() are
                # in milliseconds so we need to divide them by 1000 here.
                month = datetime.utcfromtimestamp(item[0] / 1000).month
                months.append(calendar.month_name[month])
        submitted = [item[1] for item in submitted]
        accepted = [item[1] for item in accepted]
        counts = zip(submitted, accepted)
        teams_data.append([team, counts])
    context['teams_data'] = teams_data
    context['months'] = months
    return render_to_response('patchmetrics/teams.html', context)


def faq(request):
    context = PatchworkRequestContext(request)
    context['ignored_projects'] = Project.objects.filter(
        linkname__in=PROJECTS_TO_IGNORE_IN_STATS)
    return render_to_response('patchmetrics/faq.html', context)


def frontpage(request):
    context = PatchworkRequestContext(request)
    one_month_ago = datetime.now() - relativedelta.relativedelta(months=1)
    year = one_month_ago.year
    month = one_month_ago.month
    total, accepted = get_submitted_patch_counts_for_month(year, month)
    total_waiting, old_waiting = get_number_of_patches_waiting_for_feedback()
    projects = Project.objects.exclude(
        linkname__in=PROJECTS_TO_IGNORE_IN_STATS)
    # rss_data = Feed2JSON().get_cached_data(RSS_URL)
    # Remove the feed from the front page, need also to clear feed_data dir.
    rss_data = {}
    highlight = None
    if 'entries' in rss_data and len(rss_data['entries']) > 0:
        highlight = rss_data['entries'][0]
        context['monthly_highlight_title'] = highlight['title']
        context['monthly_highlight_post'] = highlight['content'][0]['value']
    context['old_patches'] = get_oldest_patches_waiting_for_feedback()
    context['latest_patches'] = LinaroPatch.objects.order_by('-date')[:10]
    context['projects'] = projects.extra(
        select={'lower_name': 'lower(name)'}).order_by('lower_name')
    context['teams'] = Team.objects.all().order_by('display_name')
    context['month'] = calendar.month_name[month]
    context['distinct_monthly_authors'] = get_distinct_author_count(
        year, month)
    context['total_accepted'] = get_number_of_patches_accepted()
    context['submitted_on_month'] = total
    context['submitted_and_accepted_on_month'] = accepted
    context['waiting_for_feedback'] = total_waiting
    context['more_than_30_days_old_waiting_for_feedback'] = old_waiting
    context['old_patches_accepted_on_month'] = (
        get_number_of_old_patches_accepted_on_month(year, month))
    context['patches_per_month_chart'] = get_overall_patches_per_month_chart(
        year, month)
    age_increment = 15 # days
    new = get_patches_waiting_for_feedback_by_age_range(age_increment)
    context['patch_age_distribution_chart'] = (
        get_patch_age_distribution_chart(new, age_increment, 'patchAgeChart'))
    # Use a shorter age increment for the chart below.
    age_increment = 7 # days
    accepted = get_patches_accepted_by_age_range(age_increment=age_increment)
    context['time_to_acceptance_chart'] = (
        get_patch_age_distribution_chart(
            accepted, age_increment, 'timeToAcceptanceChart',
            color="'#89b529'"))
    return render_to_response('patchmetrics/frontpage.html', context)


def get_patch_age_distribution_chart(data, age_increment, div_id, color=0):
    # If we have too many ticks, the chart becomes unreadable, so we impose a
    # limit on the number of ticks and group the remaining ones together.
    ticks_limit = 6
    # Must cast the counts to int because they might come as long from the DB
    # and Javascript doesn't understand Python's long representation.
    data = sorted([[age_range, int(count)] for (age_range, count) in data])
    patches = data[:ticks_limit]
    patches_to_group = data[ticks_limit:]
    ticks = [item[0] for item in patches]
    tick_labels = {}
    for tick in ticks:
        start = tick * age_increment
        tick_labels[tick] = "%s-%s" % (start, start + age_increment)

    if len(patches_to_group) > 0:
        # If we have more than the defined limit of ticks, we sum all the
        # counts together and present them with a distinct label to indicate
        # the grouping.
        grouped_total = sum(count for (age_range, count) in patches_to_group)
        age_range = patches_to_group[0][0]
        ticks.append(age_range)
        patches.append([age_range, grouped_total])
        tick_labels[age_range] = "%s+ days" % (age_range * age_increment)

    html = r"""
        <script language="JavaScript" type="text/javascript">
            $(function () {
                var labels = %(tick_labels)s;
                function tickFormatter(val, axis) {
                    return labels[val];
                }
                var options = {
                    series: {bars: {show: true}, points: {show: true}},
                    valueLabels: {show: true},
                    bars: {barWidth: 0.5},
                    xaxis: {ticks: %(ticks)s, tickFormatter: tickFormatter}};
                var data = [{data: %(patches)s, color: %(color)s}]
                $.plot($("#%(div_id)s"), data, options);
            });
        </script>
        <div id="%(div_id)s" style="width: 350px; height: 250px"></div>
        """ % dict(ticks=ticks, patches=patches, div_id=div_id,
                   tick_labels=tick_labels, age_increment=age_increment,
                   color=color)
    return html


def get_overall_patches_per_month_chart(year, month):
    submitted, accepted = get_patches_per_month_up_to(
        year, month, get_overall_patch_count, limit=8)
    ticks = [item[0] for item in submitted]
    return generate_patches_per_month_chart(ticks, submitted, accepted)


def generate_patches_per_month_chart(ticks, submitted, accepted, width=350,
                                     height=250):
    html = """
        <script language="JavaScript" type="text/javascript">
            $(function () {
                var options = {
                    series: {lines: {show: true}, points: {show: true}},
                    legend: {position: "nw"},
                    valueLabels: {show: true},
                    xaxis: {mode: "time", timeformat: "%%y-%%m",
                            ticks: %(ticks)s}};
                $.plot(
                    $("#perMonthChart"),
                    [{label: "Submitted", data: %(submitted)s},
                     {label: "Accepted", color: "#89b529",
                      data: %(accepted)s}],
                    options);
            });
        </script>
        <div id="perMonthChart"
             style="width: %(width)dpx; height: %(height)dpx"></div>
        """ % dict(ticks=ticks, submitted=submitted, accepted=accepted,
                   width=width, height=height)
    return html


def get_patches_per_month_chart(project_or_team, year, month):
    submitted, accepted = get_patches_per_month_up_to(
        year, month, project_or_team.get_patch_count, limit=8)
    ticks = [item[0] for item in submitted]
    return generate_patches_per_month_chart(
        ticks, submitted, accepted, width=500, height=250)


def non_linaro_patches(request):
    patches = Patch.objects.exclude(
        author__in=Person.objects.filter(email__iendswith='@linaro.org'))
    projects_to_skip = Project.objects.filter(
        linkname__in=PROJECTS_TO_IGNORE_IN_STATS)
    patches = patches.exclude(
        models.Q(state=State.objects.get(name='Superseded'))
        | models.Q(project__in=projects_to_skip))

    # XXX: This is a hack because generic_list() is not meant to be used for
    # cross-project patch lists.
    context = generic_list(
        request, None, 'patchmetrics.views.non_linaro_patches',
        patches=patches)
    return render_to_response('patchmetrics/non-linaro.html', context)


def monthly_highlights(request):
    context = PatchworkRequestContext(request)
    rss_data = Feed2JSON().get_cached_data(RSS_URL)
    for entry in rss_data.get('entries'):
        entry['updated_datetime'] = datetime.fromtimestamp(
            mktime(entry["updated_parsed"]))
        entry['post'] = entry['content'][0]['value']
    return render_to_response(
        'patchmetrics/monthly-highlights.html', rss_data, context)
