#!/usr/bin/python
#
# Patchwork - automated patch tracking system
# Copyright (C) 2008 Jeremy Kerr <jk@ozlabs.org>
#
# This file is part of the Patchwork package.
#
# Patchwork is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Patchwork is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Patchwork; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import _pythonpath
import sys
import re
import chardet
import datetime
import time
import operator
import settings
from email import message_from_file
try:
    from email.header import Header, decode_header
    from email.utils import parsedate_tz, mktime_tz
except ImportError:
    # Python 2.4 compatibility
    from email.Header import Header, decode_header
    from email.Utils import parsedate_tz, mktime_tz

from patchwork.parser import parse_patch
from patchwork.models import Patch, Project, Person, Comment, State
from django.contrib.auth.models import User

default_patch_state = 'New'
list_id_headers = ['List-ID', 'X-Mailing-List', 'X-list']

whitespace_re = re.compile('\s+')
def normalise_space(str):
    return whitespace_re.sub(' ', str).strip()

def clean_header(header):
    """ Decode (possibly non-ascii) headers """

    def decode(fragment):
        (frag_str, frag_encoding) = fragment
        if frag_encoding:
            return frag_str.decode(frag_encoding)
        return frag_str.decode()

    fragments = map(decode, decode_header(header))

    return normalise_space(u' '.join(fragments))

def find_project_by_listid(mail):
    project = None
    listid_res = [re.compile('.*<([^>]+)>.*', re.S),
                  re.compile('^([\S]+)$', re.S)]

    for header in list_id_headers:
        if header in mail:

            for listid_re in listid_res:
                match = listid_re.match(mail.get(header))
                if match:
                    break

            if not match:
                continue

            listid = match.group(1)

            try:
                project = Project.objects.get(listid = listid)
                break
            except:
                pass

    return project

def extract_email_addresses(str):
    email_re = re.compile(
        r"([\w\.\-+=]+@(?:(?:[0-9a-zA-Z-]{1,}\.)*)[a-zA-Z]{2,})")
    return email_re.findall(str)

def find_project_by_list_address(mail):
    recipients = mail.get('To', '') + ' ' + mail.get('CC', '')
    email_addresses = extract_email_addresses(recipients)
    for email_address in email_addresses:
        try:
            return Project.objects.get(listemail = email_address)
        except Project.DoesNotExist:
            pass
    print ("Unable to find a project for any of the recipients: %s" %
           email_addresses)
    return None

def find_project(mail):
    project = find_project_by_listid(mail)
    if project is None and settings.PATCHWORK_FALLBACK_TO_LISTEMAIL:
        project = find_project_by_list_address(mail)
    return project

def parse_name_and_email(text):
    # tuple of (regex, fn)
    #  - where fn returns a (name, email) tuple from the match groups resulting
    #    from re.match().groups()
    from_res = [
        # for "Firstname Lastname" <example@example.com> style addresses
       (re.compile('"?(.*?)"?\s*<([^>]+)>'), (lambda g: (g[0], g[1]))),

       # for example@example.com (Firstname Lastname) style addresses
       (re.compile('"?(.*?)"?\s*\(([^\)]+)\)'), (lambda g: (g[1], g[0]))),

       # everything else
       (re.compile('(.*)'), (lambda g: (None, g[0]))),
    ]

    for regex, fn in from_res:
        match = regex.match(text)
        if match:
            name, email = fn(match.groups())
            if name is not None:
                name = name.strip()
            if email is not None:
                email = email.strip()
            return name, email
    return None, None


def find_submitter(mail):

    from_header = clean_header(mail.get('From'))
    name, email = parse_name_and_email(from_header)

    if email is None:
        raise Exception("Could not parse From: header")

    new_person = False
    try:
        person = Person.objects.get(email__iexact = email)
    except Person.DoesNotExist:
        person = Person(name = name, email = email)
        new_person = True

    return (person, new_person)

def mail_date(mail):
    t = parsedate_tz(mail.get('Date', ''))
    if not t:
        return datetime.datetime.utcnow()
    return datetime.datetime.utcfromtimestamp(mktime_tz(t))

def mail_headers(mail):
    return reduce(operator.__concat__,
            ['%s: %s\n' % (k, Header(v, header_name = k, \
                    continuation_ws = '\t').encode()) \
                for (k, v) in mail.items()])

def find_pull_request(content):
    git_re = re.compile('^The following changes since commit.*' +
                        '^are available in the git repository at:\n'
                        '^\s*([\S]+://[^\n]+)$',
                           re.DOTALL | re.MULTILINE)
    match = git_re.search(content)
    if match:
        return match.group(1)
    return None

def string_to_unicode(text, charset=None):
    """Converts the given string to unicode.

    Will first try the character set passed in and if that fails, will use
    chardet to detect a probable encoding. During this second decode the
    decoder is set to replace characters that don't decode properly with
    the UTF8 [?] character

    """
    # if we don't have a charset, detect it
    if charset is None:
        charset = chardet.detect(text)['encoding']

    if text is None or charset is None:
        return None  # Nothing to decode / no character encoding makes sense.

    if not isinstance(text, unicode):
        try:
            text = unicode(text, charset)
        except (UnicodeDecodeError, LookupError):
            # Character set was incorrectly set or message is malformed.
            # Try detecting the character set. LookupError means charset is
            # not a character set that we understand (probably garbage).
            # UnicodeDecodeError means that charset is a valid encoding, but
            # text can not be decoded using it.
            charset = chardet.detect(text)['encoding']

            if not charset:
                return None  # chardet can't work out a character set.

            # Sometimes chardet says that the encoding is <something> and it
            # doesn't decode (input charset may agree). This is normally
            # an illegally encoded email - most I have seen are gb2312 (Chinese
            # simplified). Using errors="replace" works fine and replaces
            # unknown characters with [?]. This should make it obvious to
            # anyone viewing the output if the decode worked well or not.
            text = unicode(text, charset, errors="replace")

    return text

def find_content(project, mail):
    patchbuf = None
    commentbuf = ''
    pullurl = None

    for part in mail.walk():
        if part.get_content_maintype() != 'text':
            continue

        payload = part.get_payload(decode=True)
        charset = part.get_content_charset()
        subtype = part.get_content_subtype()

        payload = string_to_unicode(payload, charset)

        if subtype in ['x-patch', 'x-diff']:
            patchbuf = payload

        elif subtype == 'plain':
            c = payload

            if not patchbuf:
                (patchbuf, c) = parse_patch(payload)

            if not pullurl:
                pullurl = find_pull_request(payload)

            if c is not None:
                commentbuf += c.strip() + '\n'

    patch = None
    comment = None

    if pullurl or patchbuf:
        name = clean_subject(mail.get('Subject'), [project.linkname])
        patch = Patch(name = name, pull_url = pullurl, content = patchbuf,
                    date = mail_date(mail), headers = mail_headers(mail))

    if commentbuf:
        if patch:
            cpatch = patch
        else:
            cpatch = find_patch_for_comment(project, mail)
            if not cpatch:
                return (None, None)
        comment = Comment(patch = cpatch, date = mail_date(mail),
                content = clean_content(commentbuf),
                headers = mail_headers(mail))

    return (patch, comment)

def find_patch_name_and_author(comment):
    """Return the patch name and author as specified in the given comment.

    The name and author of a patch are sometimes different from the Subject
    and From header fields and so are specified in the body of the message,
    before the actual patch. This is identified by "From: " and "Subject: "
    lines starting the body, which is the format understood by 'git-am'.
    """
    if not comment:
        return None, None
    lines = comment.content.split('\n')
    if len(lines) == 0:
        return None, None
    # Append an extra line at the end so that our code can assume there will
    # always be at least two lines, simplifying things significantly.
    lines.append('')
    first_line, second_line = lines[:2]
    author_line = subject_line = None
    if first_line.startswith('From:'):
        author_line = first_line
        if second_line.startswith('Subject:'):
            subject_line = second_line
    elif first_line.startswith('Subject:'):
        subject_line = first_line
        if second_line.startswith('From:'):
            author_line = second_line
    else:
        return None, None

    subject = None
    if subject_line is not None:
        subject = subject_line.replace('Subject:', '').strip()

    author = None
    if author_line is None:
        return subject, None

    name, email = parse_name_and_email(author_line.replace('From:', ''))
    if email is None:
        return subject, None

    try:
        person = Person.objects.get(email__iexact=email)
    except Person.DoesNotExist:
        person = Person(name=name, email=email)
        person.save()
    return subject, person

def find_patch_for_comment(project, mail):
    # construct a list of possible reply message ids
    refs = []
    if 'In-Reply-To' in mail:
        refs.append(mail.get('In-Reply-To'))

    if 'References' in mail:
        rs = mail.get('References').split()
        rs.reverse()
        for r in rs:
            if r not in refs:
                refs.append(r)

    for ref in refs:
        patch = None

        # first, check for a direct reply
        try:
            patch = Patch.objects.get(project = project, msgid = ref)
            return patch
        except Patch.DoesNotExist:
            pass

        # see if we have comments that refer to a patch
        try:
            comment = Comment.objects.get(patch__project = project, msgid = ref)
            return comment.patch
        except Comment.DoesNotExist:
            pass


    return None

split_re = re.compile('[,\s]+')

def split_prefixes(prefix):
    """ Turn a prefix string into a list of prefix tokens

    >>> split_prefixes('PATCH')
    ['PATCH']
    >>> split_prefixes('PATCH,RFC')
    ['PATCH', 'RFC']
    >>> split_prefixes('')
    []
    >>> split_prefixes('PATCH,')
    ['PATCH']
    >>> split_prefixes('PATCH ')
    ['PATCH']
    >>> split_prefixes('PATCH,RFC')
    ['PATCH', 'RFC']
    >>> split_prefixes('PATCH 1/2')
    ['PATCH', '1/2']
    """
    matches = split_re.split(prefix)
    return [ s for s in matches if s != '' ]

re_re = re.compile('^(re|fwd?)[:\s]\s*', re.I)
prefix_re = re.compile('^\[([^\]]*)\]\s*(.*)$')

def clean_subject(subject, drop_prefixes = None):
    """ Clean a Subject: header from an incoming patch.

    Removes Re: and Fwd: strings, as well as [PATCH]-style prefixes. By
    default, only [PATCH] is removed, and we keep any other bracketed data
    in the subject. If drop_prefixes is provided, remove those too,
    comparing case-insensitively.

    >>> clean_subject('meep')
    'meep'
    >>> clean_subject('Re: meep')
    'meep'
    >>> clean_subject('[PATCH] meep')
    'meep'
    >>> clean_subject('[PATCH] meep \\n meep')
    'meep meep'
    >>> clean_subject('[PATCH RFC] meep')
    '[RFC] meep'
    >>> clean_subject('[PATCH,RFC] meep')
    '[RFC] meep'
    >>> clean_subject('[PATCH,1/2] meep')
    '[1/2] meep'
    >>> clean_subject('[PATCH RFC 1/2] meep')
    '[RFC,1/2] meep'
    >>> clean_subject('[PATCH] [RFC] meep')
    '[RFC] meep'
    >>> clean_subject('[PATCH] [RFC,1/2] meep')
    '[RFC,1/2] meep'
    >>> clean_subject('[PATCH] [RFC] [1/2] meep')
    '[RFC,1/2] meep'
    >>> clean_subject('[PATCH] rewrite [a-z] regexes')
    'rewrite [a-z] regexes'
    >>> clean_subject('[PATCH] [RFC] rewrite [a-z] regexes')
    '[RFC] rewrite [a-z] regexes'
    >>> clean_subject('[foo] [bar] meep', ['foo'])
    '[bar] meep'
    >>> clean_subject('[FOO] [bar] meep', ['foo'])
    '[bar] meep'
    """

    subject = clean_header(subject)

    if drop_prefixes is None:
        drop_prefixes = []
    else:
        drop_prefixes = [ s.lower() for s in drop_prefixes ]

    drop_prefixes.append('patch')

    # remove Re:, Fwd:, etc
    subject = re_re.sub(' ', subject)

    subject = normalise_space(subject)

    prefixes = []

    match = prefix_re.match(subject)

    while match:
        prefix_str = match.group(1)
        prefixes += [ p for p in split_prefixes(prefix_str) \
                        if p.lower() not in drop_prefixes]

        subject = match.group(2)
        match = prefix_re.match(subject)

    subject = normalise_space(subject)

    subject = subject.strip()
    if prefixes:
        subject = '[%s] %s' % (','.join(prefixes), subject)

    return subject

sig_re = re.compile('^(-- |_+)\n.*', re.S | re.M)
def clean_content(str):
    """ Try to remove signature (-- ) and list footer (_____) cruft """
    str = sig_re.sub('', str)
    return str.strip()

def get_state(state_name):
    """ Return the state with the given name or the default State """
    if state_name:
        try:
            return State.objects.get(name__iexact=state_name)
        except State.DoesNotExist:
            pass
    return State.objects.get(name=default_patch_state)

def get_delegate(delegate_email):
    """ Return the delegate with the given email or None """
    if delegate_email:
        try:
            return User.objects.get(email__iexact=delegate_email)
        except User.DoesNotExist:
            pass
    return None

def parse_mail(mail):

    # some basic sanity checks
    if 'From' not in mail:
        return 0

    if 'Subject' not in mail:
        return 0

    if 'Message-Id' not in mail:
        return 0

    hint = mail.get('X-Patchwork-Hint', '').lower()
    if hint == 'ignore':
        return 0;

    project = find_project(mail)
    if project is None:
        print "no project found"
        return 0

    msgid = mail.get('Message-Id').strip()

    (submitter, save_required) = find_submitter(mail)

    (patch, comment) = find_content(project, mail)

    if patch:
        # we delay the saving until we know we have a patch.
        if save_required:
            submitter.save()
            save_required = False
        patch.submitter = submitter
        name, author = find_patch_name_and_author(comment)
        if author is not None:
            patch.author = author
        if name is not None:
            patch.name = name
        patch.msgid = msgid
        patch.project = project
        patch.state = get_state(mail.get('X-Patchwork-State', '').strip())
        patch.delegate = get_delegate(
                mail.get('X-Patchwork-Delegate', '').strip())
        try:
            patch.save()
        except Exception, ex:
            print str(ex)

    if comment:
        if save_required:
            submitter.save()
        # looks like the original constructor for Comment takes the pk
        # when the Comment is created. reset it here.
        if patch:
            comment.patch = patch
        comment.submitter = submitter
        comment.msgid = msgid
        try:
            comment.save()
        except Exception, ex:
            print str(ex)

    return 0

def main(args):
    mail = message_from_file(sys.stdin)
    return parse_mail(mail)

if __name__ == '__main__':
    sys.exit(main(sys.argv))
