#!/usr/bin/python

# XXX: Currently we have support for only one source_tree per project, which
# means a patch will be automatically marked as accepted only when it's found
# on that source tree. This is not an ideal representation of projects like
# u-boot and the kernel, where patches go into sub-maintainer/custodian
# branches once they are accepted and only after lingering there for quite
# some time they're pulled into the master branch, and it means patches
# accepted into sub-maintainer/custodian branches will only be marked as
# accepted once they reach the master branch.

import _pythonpath
import sys
from datetime import datetime
from operator import attrgetter

from lockfile import FileLock, LockFailed, LockTimeout

from django.db.models import Q
from dulwich.repo import Repo

from patchwork.models import Project, State
from patchwork.utils import (
    ensure_source_checkout_for_project,
    get_commits_to_parse,
    get_diff_for_commit,
    get_patches_matching_commit,
    GitException
    )


def main():
    projects = Project.objects.exclude(
        Q(source_tree__isnull=True) | Q(source_tree=''))

    print "Total number of projects: {0}".format(len(projects))

    for project in projects:
        print "=" * 80
        print "Scanning commits of {0}".format(project.name)

        try:
            root = ensure_source_checkout_for_project(project)
        except GitException, e:
            sys.stderr.write(str(e) + "\n")
            continue

        patches_accepted = 0
        patches_superseded = 0

        print "Last commit analyzed: {0}".format(project.last_seen_commit_ref)

        try:
            commit_ids = get_commits_to_parse(
                root, start_at=project.last_seen_commit_ref)
        except GitException, e:
            print >> sys.stderr, (
                "Project {0}: {1}\nLast commit seen: {2}".format(
                    project.name, str(e), project.last_seen_commit_ref))
            print "=" * 80, "\n"
            continue

        print "Total commits to analyze: {0}".format(len(commit_ids))

        repo = Repo(root)
        for commit_id in commit_ids:
            project.last_seen_commit_ref = commit_id
            project.save()
            diff = get_diff_for_commit(root, commit_id)
            if diff is None:
                # Some commits result in empty diffs (e.g. only changing file
                # permissions); in these cases we just skip them.
                continue
            commit = repo[commit_id]
            patches = get_patches_matching_commit(
                project, diff, commit.message, commit.author)
            if len(patches) == 0:
                continue
            elif len(patches) == 1:
                matching_patch = patches[0]
            else:
                # There are multiple versions of this patch, so we'll mark the
                # latest one as accepted and all others as superseded.
                patches = sorted(patches, key=attrgetter('date'))
                matching_patch = patches.pop()
                for patch in patches:
                    patches_superseded += 1
                    patch.state = State.objects.get(name='Superseded')
                    patch.save()

            matching_patch.state = State.objects.get(name='Accepted')
            matching_patch.commit_ref = commit.id
            matching_patch.save()
            # TODO: Maybe we need a new timestamp field to store the date when
            # a patch is accepted upstream? It may be fine to use
            # date_last_state_change for that, but it's slightly weird that
            # here we set it to a date in the past which doesn't match the
            # date of the state change in patchwork but probably matches the
            # date of the state change in the real world.
            date = datetime.utcfromtimestamp(commit.commit_time)
            if date >= matching_patch.date:
                matching_patch.date_last_state_change = date
            else:
                # This is probably because the patch arrived to us after it
                # was committed. It could be that the email was delayed or
                # something caused a delay in our processing of it. In this
                # case we don't want to have the date_last_state_change before
                # the patch's creation, so we set it to the patch's creation
                # date.
                matching_patch.date_last_state_change = matching_patch.date
            matching_patch.save()
            patches_accepted += 1

        print "{0} patches marked as accepted".format(patches_accepted)
        print "{0} patches marked as superseded".format(patches_superseded)
        print "=" * 80, "\n"


lock = FileLock('/var/lock/update-committed-patches')
try:
    lock.acquire(timeout=10)
except LockFailed, e:
    print >> sys.stderr, "Failed to acquire lock: {0}\n".format(e)
except LockTimeout:
    print >> sys.stderr, ("Another process seems to be running; "
                          "can't acquire lock.\n")
else:
    try:
        main()
    finally:
        lock.release()
