
from django.db.utils import IntegrityError
from django.test import TestCase

from patchwork.models import Person, State
from patchwork.tests.factory import factory


class UserProfileTestCase(TestCase):

    def test_submitted_patches_waiting_feedback(self):
        # Create two people linked to the same user.
        person = factory.makePerson(is_user=True)
        profile = person.user.get_profile()
        person2 = factory.makePerson(is_user=False)
        person2.user = person.user
        person2.save()

        # Create 3 patches that
        #   - apply to the same project;
        #   - have been submitted by a Person linked to our newly created user
        # Of those 3 patches, only the first two are in a state that needs
        # action, so only those will be returned by
        # submitted_patches_waiting_feedback().
        # We also create a couple more patches, but one of them applies to a
        # different project than the ones created above and the other has not
        # been submitted by a Person linked to our user, so they are not in
        # the patches returned by submitted_patches_waiting_feedback().
        patch1 = factory.makePatch(submitter=person)
        project = patch1.project
        patch2 = factory.makePatch(project=project, submitter=person2)
        patch3 = factory.makePatch(project=project, submitter=person2)
        patch3.state = State.objects.get(name='Accepted')
        patch3.save()
        patch4 = factory.makePatch(submitter=person2)
        patch5 = factory.makePatch(project=project)

        # Here we see that UserProfile.submitted_patches_waiting_feedback()
        # only returns the two patches that are in a state that requires
        # action and that have been submitted by a person linked to that
        # profile.
        patches = profile.submitted_patches_waiting_feedback(patch1.project)
        self.assertEquals([patch1, patch2], list(patches.order_by('id')))


class PatchTestCase(TestCase):

    def test_date_last_state_change(self):
        patch = factory.makePatch()
        self.assertEqual(None, patch.date_last_state_change)
        patch.state = State.objects.get(name='Accepted')
        patch.save()
        self.assertNotEqual(None, patch.date_last_state_change)


class PersonTestCase(TestCase):

    def test_email_uniqueness_is_case_insensitive(self):
        Person(email='foo.bar@example.com').save()
        p2 = Person(email='Foo.Bar@example.com')
        self.assertRaises(IntegrityError, p2.save)
