import atexit
import multiprocessing
import os
import shutil
import subprocess
import tempfile
import textwrap
from multiprocessing.sharedctypes import Array, Value
from unittest import TestCase

from dulwich.repo import Repo

from patchwork.utils import (
    get_child_pids, get_commits_to_parse, get_diff_for_commit,
    get_patches_matching_commit)
from patchwork.tests.factory import factory


class TestCommitToPatchMatcher(TestCase):
    """Test get_patches_matching_commit()."""

    def test_identical_diffs(self):
        patch = factory.makePatch()
        commit_diff = patch.content
        # If the commit diff is identical to that of our patch, the commit
        # message and author are irrelevant.
        self.assertEqual([patch],
                         get_patches_matching_commit(
                             patch.project, commit_diff, 'msg', 'author'))

    def test_no_matching_patch(self):
        patch = factory.makePatch()
        self.assertEqual([],
                         get_patches_matching_commit(
                             patch.project, 'not-a-diff', 'msg', 'author'))

    def test_identical_title_and_author(self):
        # When the commit's title (first line of the message) and author are
        # identical a the patch's name and author, it'll be a match even if
        # the diff is not identical.
        patch = factory.makePatch()
        commit_diff = patch.content + "\n+An extra line at the bottom"
        author = "%s <%s>" % (patch.author.name, patch.author.email)
        self.assertEqual([patch],
                         get_patches_matching_commit(
                             patch.project, commit_diff, patch.name, author))

    def test_identical_author_with_similar_title_and_diff(self):
        # Here we get a match because the author is the same with the diff
        # and commit title being only slightly different.
        patch = factory.makePatch()
        commit_diff = patch.content + "\n+An extra line at the bottom"
        author = "%s <%s>" % (patch.author.name, patch.author.email)
        self.assertEqual([patch],
                         get_patches_matching_commit(
                             patch.project, commit_diff, patch.name + 'a',
                             author))


class TestGitRepoScanning(TestCase):
    """Tests for helper functions that scan commits on a git repo."""

    def test_get_commits_to_parse(self):
        repo = create_git_repo()
        commit, commit2, commit3 = self.create_commits(repo)
        self.assertEqual(
            [commit, commit2, commit3],
            get_commits_to_parse(repo, start_at=None))

    def test_get_commits_to_parse_starting_at_specific_commit(self):
        repo = create_git_repo()
        commit, commit2, commit3 = self.create_commits(repo)
        self.assertEqual(
            [commit3], get_commits_to_parse(repo, start_at=commit2))

    def test_get_diff_for_commit(self):
        repo = create_git_repo()
        commit = add_file_and_commit(repo, 'foobar', 'FooBar')
        expected_diff = textwrap.dedent("""\
            diff --git a/foobar b/foobar
            new file mode 100644
            index 0000000..61e2c47
            --- /dev/null
            +++ b/foobar
            @@ -0,0 +1 @@
            +FooBar
            """)
        self.assertEqual(expected_diff, get_diff_for_commit(repo, commit))

    def test_get_diff_for_empty_commit(self):
        repo = create_git_repo()
        empty_commit = add_file_and_commit(repo, 'foobar', '')
        # This commit doesn't have a diff as it just adds a file with no
        # content, so get_diff_for_commit() returns None.
        self.assertEqual(None, get_diff_for_commit(repo, empty_commit))

    def create_commits(self, repo):
        # Create 3 commits on the given repo and return their IDs.  Each of
        # these commits just add a single file with some arbitrary data in it.
        commit = add_file_and_commit(repo, 'foo', 'Content1')
        commit2 = add_file_and_commit(repo, 'bar', 'Content2')
        commit3 = add_file_and_commit(repo, 'baz', 'Content3')
        return commit, commit2, commit3


class TestGetChildPIDs(TestCase):

    def test_get_child_pids(self):
        """Fire two nested subprocesses to test get_child_pids().

        We start a new subprocess using the multiprocessing module and that
        subprocess, in turn, will start another one.

        Then before we join those subprocesses we call get_child_pids() to get
        the PIDs of the two children of this process. At that point we also
        store the PID of the grandchild process to later compare against the
        return value of get_child_pids().
        """
        parent = Value('i', os.getpid())
        grandchild_pid = Value('i')
        # Even though we only expect get_child_pids() to return 2 PIDs,
        # allocate an array with length==10 so that if there's a bug which
        # causes get_child_pids() to return more than two PIDs we don't cause
        # the function below to write across the boundaries, which would make
        # the test crash.
        observed_pids = Array('i', [-1]*10)
        def target(parent, grandchild_pid, observed_pids):
            # Here we fire the last subprocess and before join()ing we call
            # get_child_pids() passing the PID of our parent.
            grandchild = multiprocessing.Process(target=lambda: None)
            grandchild.start()
            grandchild_pid.value = grandchild.pid
            child_pids = get_child_pids(parent.value)
            for i in range(len(child_pids)):
                observed_pids[i] = int(child_pids[i])
            grandchild.join()
        child = multiprocessing.Process(
            target=target, args=(parent, grandchild_pid, observed_pids))
        child.start()
        child.join()
        self.assertEqual(
            sorted([pid for pid in observed_pids[:] if pid != -1]),
            sorted([child.pid, grandchild_pid.value]))


def run_git_in_subprocess(root, cmd):
    proc = subprocess.Popen(
        cmd, cwd=root, stdout=subprocess.PIPE, shell=True)
    out, err = proc.communicate()
    assert proc.returncode == 0, (cmd, out, err)
    return out, err


def create_git_repo():
    tmpdir = tempfile.mkdtemp()
    atexit.register(shutil.rmtree, tmpdir)
    run_git_in_subprocess(tmpdir, 'git init .')
    return tmpdir


def create_arbitrary_commit():
    root = create_git_repo()
    commit_id = add_file_and_commit(root, 'foo', 'Bar')
    return Repo(root)[commit_id]


def add_file_and_commit(root, filename, data, message='foo', author=None):
    with open(os.path.join(root, filename), 'w') as fd:
        fd.write(data)
    run_git_in_subprocess(root, 'git add %s' % filename)
    cmd = 'git commit -m "%s"' % message
    if author is not None:
        cmd += ' --author="%s"' % author
    run_git_in_subprocess(root, cmd)
    out, err = run_git_in_subprocess(root, 'git rev-parse HEAD')
    return out.strip()
