# Patchwork - automated patch tracking system
# Copyright (C) 2008 Jeremy Kerr <jk@ozlabs.org>
#
# This file is part of the Patchwork package.
#
# Patchwork is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Patchwork is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Patchwork; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import atexit
import chardet
import os
import re
import subprocess
import time

from difflib import SequenceMatcher
from django.conf import settings
from django.shortcuts import get_object_or_404
from patchwork.bin.parsemail import parse_name_and_email
from patchwork.models import Bundle, BundlePatch, Patch, Person, State
from patchwork.parser import hash_patch, parse_patch


def close_fo(fo):
    """Very simple function to close a file object.

    :param fo: The file object to close.
    """
    fo.close()

# Timeout for git operations.
GIT_TIMEOUT = 120 * 60
# FIXME This is a temporary solution, a workardound for bug 1017933.
# Once that bug is fixed or in the process, this should be removed.
REPO_BRANCHES = {'cpufreq': 'origin/next'}

DEVNULL = open(os.devnull, "w")
atexit.register(close_fo, DEVNULL)


class GitException(Exception):
    """General class for git exceptions."""


class FailedToFetchGitRepo(GitException):
    """Failed to fetch the git repo of a project."""


class FailedToCheckoutGitRepo(GitException):
    """Failed to perform a checkout operation on a git repository."""


class GitTimeout(GitException):
    """Timeout when fetching the git repo of a project."""


def get_patch_ids(d, prefix='patch_id'):
    ids = []

    for (k, v) in d.items():
        a = k.split(':')
        if len(a) != 2:
            continue
        if a[0] != prefix:
            continue
        if not v:
            continue
        ids.append(a[1])

    return ids


class Order(object):
    order_map = {
        'date':         'date',
        'name':         'name',
        'state':        'state__ordering',
        'submitter':    'submitter__name',
        'delegate':     'delegate__username',
    }
    default_order = ('date', True)

    def __init__(self, str=None, editable=False):
        self.reversed = False
        self.editable = editable
        (self.order, self.reversed) = self.default_order

        if self.editable:
            return

        if str is None or str == '':
            return

        reversed = False
        if str[0] == '-':
            str = str[1:]
            reversed = True

        if str not in self.order_map.keys():
            return

        self.order = str
        self.reversed = reversed

    def __str__(self):
        str = self.order
        if self.reversed:
            str = '-' + str
        return str

    def name(self):
        return self.order

    def reversed_name(self):
        if self.reversed:
            return self.order
        else:
            return '-' + self.order

    def query(self):
        q = self.order_map[self.order]
        if self.reversed:
            q = '-' + q
        return q


bundle_actions = ['create', 'add', 'remove']


def set_bundle(user, project, action, data, patches, context):
    # set up the bundle
    bundle = None
    if action == 'create':
        bundle_name = data['bundle_name'].strip()
        if not bundle_name:
            return ['No bundle name was specified']

        bundle = Bundle(owner=user, project=project, name=bundle_name)
        bundle.save()
        context.add_message("Bundle {0} created".format(bundle.name))
    elif action == 'add':
        bundle = get_object_or_404(Bundle, id=data['bundle_id'])
    elif action == 'remove':
        bundle = get_object_or_404(Bundle, id=data['removed_bundle_id'])

    if not bundle:
        return ['no such bundle']

    for patch in patches:
        if action == 'create' or action == 'add':
            bundlepatch_count = BundlePatch.objects.filter(bundle=bundle,
                                                           patch=patch).count()
            if bundlepatch_count == 0:
                bundle.append_patch(patch)
                context.add_message("Patch '{0}' added to bundle {1}".format(
                                    patch.name, bundle.name))
            else:
                context.add_message("Patch '{0}' already in bundle {1}".format(
                                    patch.name, bundle.name))

        elif action == 'remove':
            try:
                bp = BundlePatch.objects.get(bundle=bundle, patch=patch)
                bp.delete()
                context.add_message("Patch '{0}' removed from bundle "
                                    "{1}\n".format(patch.name, bundle.name))
            except Exception:
                pass

    bundle.save()

    return []


def get_child_pids(pid):
    """Recursively get the PIDs of the given process' children."""
    pids = []
    stdout, _ = subprocess.Popen(
        ['pgrep', '-P', str(pid)], stdout=subprocess.PIPE).communicate()
    children = stdout.split()
    while children:
        pids.extend(children)
        new_children = []
        for pid in children:
            args = ['pgrep', '-P', str(pid)]
            stdout, _ = subprocess.Popen(
                args, stdout=subprocess.PIPE).communicate()
            new_children.extend(stdout.split())
        children = new_children
    return pids


# We run git as a subprocess here instead of using a library like dulwich or
# git-python because dulwich had issues when getting diffs for certain
# commits and the version of git-python that is packaged in most distros is
# not suitable, whereas its latest version has a bunch of non-packaged
# dependencies.
def ensure_source_checkout_for_project(project, timeout=GIT_TIMEOUT):
    """Ensures that we have the code for the project we are working on.

    If there is no code in the PATCHWORK_GIT_REPOS_DIR, it performs a 'clone'
    operation, otherwise we try to 'pull' the latest changes.

    :param project: the project we are working on
    :param timeout: the timeout for the git operation, defaults to GIT_TIMEOUT
    """
    root = os.path.join(settings.PATCHWORK_GIT_REPOS_DIR, project.linkname)

    if os.path.exists(root):
        # FIXME this is a temporary solution, we need to fix bug
        # https://bugs.launchpad.net/linaro-patchmetrics/+bug/1017933
        if project.name in REPO_BRANCHES:
            ensure_on_valid_branch(project, REPO_BRANCHES.get(project.name))

        print "Pulling changes from {0}".format(project.name)
        proc = subprocess.Popen(['git', 'pull'], cwd=root,
                                bufsize=4096, stdout=DEVNULL, stderr=DEVNULL)
    else:
        print "Cloning repository {0}".format(project.source_tree)
        proc = subprocess.Popen(['git', 'clone', project.source_tree, root],
                                bufsize=4096, stdout=DEVNULL, stderr=DEVNULL)

    git_process_poll(proc, project, timeout)

    if proc.returncode != 0:
        raise FailedToFetchGitRepo(
            "Failed to fetch/pull {0} ({1} - {2})".format(
                project.source_tree, project.name, project.linkname))

    return root


def ensure_on_valid_branch(project, branch, timeout=GIT_TIMEOUT):
    """Ensures that we are on a valid git branch.

    Performs a git checkout on the provided branch name, no check is done on
    the name, it has to be a valid git branch. If we are already on that branch
    nothing happens, git will return 0 as normal.

    :param project: the project we are working on
    :param branch: the name of the branch to checkout
    :param timeout: the max timeout for git operation, defaults to GIT_TIMEOUT
    """
    cwd = os.path.join(settings.PATCHWORK_GIT_REPOS_DIR, project.linkname)
    proc = subprocess.Popen(['git', 'branch'], cwd=cwd, stdout=subprocess.PIPE)
    out, _ = proc.communicate()

    # Probably not the best way to get the branch name out of the result of
    # the process output. The `git branch` command returns something like this:
    # "* branch\n"
    # We are interested only in the "branch" part.
    name = out.strip().split()[1]
    if name in branch:
        # We already are in the correct branch/checkout.
        return

    proc = subprocess.Popen(['git', 'checkout', '--track', branch],
                            cwd=cwd, bufsize=4096, stdout=subprocess.PIPE,
                            stderr=DEVNULL)

    out, _ = proc.communicate()

    if proc.returncode != 0:
        raise FailedToCheckoutGitRepo("Failed to checkout branch '{0}' "
                                      "of {1}".format(branch, project.name))


def git_process_poll(process, project, timeout):
    """Check the status of the process.

    :param process: the process to watch
    :param project: the project we are working on
    :param timeout: the max timeout for the process
    """
    wait = 0
    wait_increment = 5

    while wait < timeout:
        if process.poll() is not None:
            break
        wait += wait_increment
        time.sleep(wait_increment)

    if process.returncode is None:
        kill_git_processes(process, project)


def kill_git_processes(process, project):
    """Kills the process of the git operation, and its subprocesses.

    :param process: the parent process to kill
    :param project: the project we are working on
    """
    cwd = os.path.join(settings.PATCHWORK_GIT_REPOS_DIR, project.linkname)
    msg = ("Timeout when performing git operation on '{0}' ({1}).\n".format(
           (project.name, project.linkname)))
    msg += (" Project path: {0}\n Project tree: {1}\n".format(
            (cwd, project.source_tree)))

    # Kill the subprocess we started and all its children. This is
    # necessary because 'git pull' will actually run a shell script
    # as a subprocess, with the script firing a 'git fetch' subprocess,
    # and if we don't kill them all individually they will be left around
    # and some times they will never finish.
    pids = [str(process.pid)]
    # We traverse the child processes recursively here instead of querying
    # all processes under the same process group because on production
    # this will be run as a cronjob and thus the process group will
    # include the shell process started by cron, which is the parent of
    # this very process and thus we don't want to kill -- we only want to
    # kill the subprocess we started above and all its children.
    pids.extend(get_child_pids(process.pid))
    msg += " Killing processes {0} and moving on.".format(pids)
    proc = subprocess.Popen(
        ['kill', '-KILL'] + pids,
        stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    kill_stdout, kill_stderr = proc.communicate()
    msg += " Output of the kill command:\n{0}\n{1}\n".format(
        kill_stdout, kill_stderr)

    # Poll the subprocess we started to ensure it's not left behind as a
    # zombie, but first sleep a couple seconds to make sure it's finished
    # when we poll.
    time.sleep(2)
    if proc.poll() is None:
        msg += ("Oops, we failed to kill the git process (pid={0}) that "
                "timed out.".format(proc.pid))
    raise GitTimeout(msg)


def get_commits_to_parse(root, start_at):
    args = ['git', 'rev-list', '--reverse']
    if start_at:
        args.append('{0}..HEAD'.format(start_at))
    else:
        args.extend(['--max-count=20000', 'HEAD'])
    proc = subprocess.Popen(args, cwd=root, stdout=subprocess.PIPE,
                            bufsize=4096)
    commit_ids = []
    while True:
        for line in iter(proc.stdout.readline, b''):
            line = line.strip()
            if line:
                commit_ids.append(line)
        if proc.poll() is not None:
            break

    if proc.returncode != 0:
        raise GitException("Error retrieving revisions list.")

    return commit_ids


def get_diff_for_commit(root, commit_id):
    proc = subprocess.Popen(
        ['git', 'show', commit_id], cwd=root, stdout=subprocess.PIPE,
        stderr=DEVNULL, bufsize=4096)

    stdout = ""
    while True:
        for line in iter(proc.stdout.readline, b''):
            stdout += line
        if proc.poll() is not None:
            break

    if proc.returncode != 0:
        print ("Error retrieving diff for commit {0}".format(commit_id))
        return None

    try:
        diff = stdout.decode('utf-8')
    except UnicodeDecodeError:
        try:
            # chardet.detect is rather slow so we only use it when we fail
            # to decode from utf-8.
            encoding = chardet.detect(stdout)['encoding']
            diff = stdout.decode(encoding)
        except UnicodeDecodeError:
            print ("Skipping commit {0} as we can't find the appropriate "
                   "charset to decode it".format(commit_id))
            return None

    patch, _ = parse_patch(diff)
    return patch


def get_patches_matching_commit(project, commit_diff, commit_msg,
                                commit_author):
    """Return the Patch objects that are likely matches to the given commit.

    Patches that are either Accepted or Superseded are never included in the
    results.

    :param project: The Project to which this patch applies.
    :param commit_diff: A string with the diff that was ultimately committed.
    :param commit_msg: The commit message, as a string.
    :param commit_author: The commit author, including the email address, also
                          as a string.
    """
    pending_patches = Patch.objects.filter(project=project).exclude(
        state__in=State.objects.filter(name__in=['Accepted', 'Superseded']))
    # First check if we have patches with a diff hash identical to that of the
    # diff that was ultimately committed.
    patches = list(pending_patches.filter(
        hash=hash_patch(commit_diff).hexdigest()))
    if len(patches) > 0:
        return patches

    # No hashes match that of the committed changes, so lookup using the
    # author's email address and the patch name.
    commit_title = commit_msg.split('\n')[0]
    _, author_email = parse_name_and_email(commit_author)
    if author_email is None:
        # No author found, nothing we can do about it.
        return []

    try:
        author = Person.objects.get(email=author_email)
    except Person.DoesNotExist:
        return []

    # Now lookup patches from the same author that have the exact same name as
    # that of the commit.
    patches = list(pending_patches.filter(name=commit_title, author=author))
    if len(patches) > 0:
        return patches

    patches = []
    for patch in pending_patches.filter(project=project, author=author):
        # Be conservative and use 0.9 as the similarity ratio between the
        # commit's title and content to prevent false-positives.
        # TODO: It'd be nice to factor this out into a separate method for
        # easier testing.
        if SequenceMatcher(None, patch.name, commit_title).ratio() > 0.9:
            matcher = SequenceMatcher(None, patch.content, commit_diff)
            if matcher.ratio() > 0.9:
                # Some commits may be just reverting a previous one and in
                # these cases the title and content ratios will be high and
                # pass the two tests above, so we do one last check here to
                # avoid false-positives in these cases.
                patch_title = re.sub(r'\[[^\]]*\]', '', patch.name).strip()
                commit_title = re.sub(r'\[[^\]]*\]', '', commit_title).strip()
                if (patch_title.startswith('Revert')
                        and not commit_title.startswith('Revert')):
                    continue
                patches.append(patch)
    return patches
